"""Constraint utility."""

__all__ = [
    "Constraint",
    "EquationConstraint",
    "InSetConstraint",
    "NotInSetConstraint",
]

from abc import ABC, abstractmethod
from typing import Any

from fhy_core.utils import Self, format_comma_separated_list

from .expression import (
    BinaryExpression,
    BinaryOperation,
    Expression,
    IdentifierExpression,
    LiteralExpression,
    LiteralType,
    copy_expression,
    pformat_expression,
    simplify_expression,
)
from .identifier import Identifier


class Constraint(ABC):
    """Abstract base class for constraints."""

    _variable: Identifier

    def __init__(self, constrained_variable: Identifier) -> None:
        self._variable = constrained_variable

    @property
    def variable(self) -> Identifier:
        return self._variable

    def __call__(self, values: dict[Identifier, Any]) -> bool:
        return self.is_satisfied(values)

    @abstractmethod
    def is_satisfied(self, variable_value: Any) -> bool:
        """Check if the value satisfies the constraint.

        Args:
            variable_value: Value to check.

        Returns:
            True if the value satisfies the constraint; False otherwise.

        """

    @abstractmethod
    def copy(self) -> Self:
        """Return a shallow copy of the constraint."""

    @abstractmethod
    def convert_to_expression(self) -> Expression:
        """Return an expression equivalent to the constraint.

        Raises:
            ValueError: If the constraint cannot be converted to an expression.

        """

    def __copy__(self) -> Self:
        return self.copy()

    @abstractmethod
    def __repr__(self) -> str: ...

    @abstractmethod
    def __str__(self) -> str: ...


class EquationConstraint(Constraint):
    """Represents an equation constraint."""

    _expression: Expression

    def __init__(
        self, constrained_variable: Identifier, expression: Expression
    ) -> None:
        super().__init__(constrained_variable)
        self._expression = expression

    def is_satisfied(self, value: Expression | LiteralType) -> bool:
        if isinstance(value, (str, float, int, bool)):
            value = LiteralExpression(value)
        result = simplify_expression(self._expression, {self.variable: value})
        return (
            isinstance(result, LiteralExpression)
            and isinstance(result.value, bool)
            and result.value
        )

    def copy(self) -> "EquationConstraint":
        new_constraint = EquationConstraint(
            self.variable, copy_expression(self._expression)
        )
        return new_constraint

    def convert_to_expression(self) -> Expression:
        return copy_expression(self._expression)

    def __repr__(self) -> str:
        return repr(self._expression)

    def __str__(self) -> str:
        return pformat_expression(self._expression)


class InSetConstraint(Constraint):
    """Represents an in-set constraint."""

    _valid_values: set[Any]

    def __init__(
        self, constrained_variable: Identifier, valid_values: set[Any]
    ) -> None:
        super().__init__(constrained_variable)
        self._valid_values = valid_values

    def is_satisfied(self, value: Any) -> bool:
        return value in self._valid_values

    def copy(self) -> "InSetConstraint":
        new_constraint = InSetConstraint(self.variable, self._valid_values.copy())
        return new_constraint

    def convert_to_expression(self) -> Expression:
        if len(self._valid_values) == 0:
            return LiteralExpression(False)
        elif len(self._valid_values) == 1:
            return self._generate_single_value_constraint(
                next(iter(self._valid_values))
            )
        else:
            return Expression.logical_or(
                *map(self._generate_single_value_constraint, self._valid_values)
            )

    def _generate_single_value_constraint(self, value: Any) -> Expression:
        if not isinstance(value, LiteralType):
            raise ValueError(
                f"Conversion of type {type(value)} to an expression is not supported."
            )
        variable = IdentifierExpression(self.variable)
        return BinaryExpression(
            BinaryOperation.EQUAL,
            variable,
            LiteralExpression(value),
        )

    def __repr__(self) -> str:
        return repr(self._valid_values)

    def __str__(self) -> str:
        return (
            f"{self.variable} in {{"
            f"{format_comma_separated_list(self._valid_values, str_func=str)}}}"
        )


class NotInSetConstraint(Constraint):
    """Represents a not-in-set constraint."""

    _invalid_values: set[Any]

    def __init__(
        self, constrained_variable: Identifier, invalid_values: set[Any]
    ) -> None:
        super().__init__(constrained_variable)
        self._invalid_values = invalid_values

    def is_satisfied(self, value: Any) -> bool:
        return value not in self._invalid_values

    def copy(self) -> "NotInSetConstraint":
        new_constraint = NotInSetConstraint(self.variable, self._invalid_values.copy())
        return new_constraint

    def convert_to_expression(self) -> Expression:
        if len(self._invalid_values) == 0:
            return LiteralExpression(True)
        elif len(self._invalid_values) == 1:
            return self._generate_single_value_constraint(
                next(iter(self._invalid_values))
            )
        else:
            return Expression.logical_and(
                *map(self._generate_single_value_constraint, self._invalid_values)
            )

    def _generate_single_value_constraint(self, value: Any) -> Expression:
        if not isinstance(value, LiteralType):
            raise ValueError(
                f"Conversion of type {type(value)} to an expression is not supported."
            )
        variable = IdentifierExpression(self.variable)
        return BinaryExpression(
            BinaryOperation.NOT_EQUAL,
            variable,
            LiteralExpression(value),
        )

    def __repr__(self) -> str:
        return repr(self._invalid_values)

    def __str__(self) -> str:
        return (
            f"{self.variable} not in {{"
            f"{format_comma_separated_list(self._invalid_values, str_func=str)}}}"
        )



"""Tests the symbol table."""

import pytest
from fhy_core.error import SymbolTableError
from fhy_core.identifier import Identifier
from fhy_core.symbol_table import (
    ImportSymbolTableFrame,
    SymbolTable,
)


@pytest.fixture
def empty_symbol_table() -> SymbolTable:
    return SymbolTable()


def test_add_and_check_namespace(empty_symbol_table: SymbolTable):
    """Test that a namespace can be added and checked."""
    namespace = Identifier("test_namespace")

    empty_symbol_table.add_namespace(namespace)

    assert empty_symbol_table.is_namespace_defined(namespace)
    assert empty_symbol_table.get_number_of_namespaces() == 1


def test_add_duplicate_namespace_fails(empty_symbol_table: SymbolTable):
    """Test that adding a duplicate namespace raises a SymbolTableError."""
    namespace = Identifier("test_namespace")

    empty_symbol_table.add_namespace(namespace)

    with pytest.raises(SymbolTableError):
        empty_symbol_table.add_namespace(namespace)


def test_get_undefined_namespace_fails(empty_symbol_table: SymbolTable):
    """Test that an undefined namespace raises a SymbolTableError."""
    undefined_namespace = Identifier("undefined_namespace")

    with pytest.raises(SymbolTableError):
        empty_symbol_table.get_namespace(undefined_namespace)


def test_add_and_get_symbol(empty_symbol_table: SymbolTable):
    """Test that a symbol can be added and retrieved."""
    namespace = Identifier("test_namespace")
    symbol_name = Identifier("test_symbol")
    frame = ImportSymbolTableFrame(symbol_name)

    empty_symbol_table.add_namespace(namespace)
    empty_symbol_table.add_symbol(namespace, symbol_name, frame)

    assert empty_symbol_table.is_symbol_defined(symbol_name)
    assert empty_symbol_table.get_frame(symbol_name) == frame


def test_add_and_get_symbol_in_namespace(empty_symbol_table: SymbolTable):
    """Test that a symbol can be added and retrieved from a namespace."""
    namespace = Identifier("test_namespace")
    symbol_name = Identifier("test_symbol")
    frame = ImportSymbolTableFrame(symbol_name)

    empty_symbol_table.add_namespace(namespace)
    empty_symbol_table.add_symbol(namespace, symbol_name, frame)

    assert empty_symbol_table.is_symbol_defined_in_namespace(namespace, symbol_name)
    assert empty_symbol_table.get_frame_from_namespace(namespace, symbol_name) == frame


def test_add_duplicate_symbol_fails(empty_symbol_table: SymbolTable):
    """Test that adding a duplicate symbol raises a SymbolTableError."""
    namespace = Identifier("test_namespace")
    symbol_name = Identifier("test_symbol")
    frame = ImportSymbolTableFrame(symbol_name)

    empty_symbol_table.add_namespace(namespace)
    empty_symbol_table.add_symbol(namespace, symbol_name, frame)

    with pytest.raises(SymbolTableError):
        empty_symbol_table.add_symbol(namespace, symbol_name, frame)


def test_get_undefined_symbol_from_namespace_fails(empty_symbol_table: SymbolTable):
    """Test that getting an undefined symbol from a namespace raises a
    SymbolTableError.
    """
    namespace = Identifier("test_namespace")
    symbol_name = Identifier("undefined_symbol")

    empty_symbol_table.add_namespace(namespace)

    with pytest.raises(SymbolTableError):
        empty_symbol_table.get_frame_from_namespace(namespace, symbol_name)


def test_add_namespace_with_parent(empty_symbol_table: SymbolTable):
    """Test that a namespace can be added with a parent namespace."""
    parent_namespace = Identifier("parent_namespace")
    child_namespace = Identifier("child_namespace")

    empty_symbol_table.add_namespace(parent_namespace)
    empty_symbol_table.add_namespace(child_namespace, parent_namespace)

    assert empty_symbol_table.is_namespace_defined(child_namespace)


def test_get_symbol_from_namespace_inherited_from_parent(
    empty_symbol_table: SymbolTable,
):
    """Test that a child namespace can access a symbol from a parent namespace."""
    parent_namespace = Identifier("parent_namespace")
    child_namespace = Identifier("child_namespace")
    symbol_name = Identifier("test_symbol")
    frame = ImportSymbolTableFrame(symbol_name)

    empty_symbol_table.add_namespace(parent_namespace)
    empty_symbol_table.add_namespace(child_namespace, parent_namespace)
    empty_symbol_table.add_symbol(parent_namespace, symbol_name, frame)

    assert empty_symbol_table.is_symbol_defined_in_namespace(
        child_namespace, symbol_name
    )
    assert (
        empty_symbol_table.get_frame_from_namespace(child_namespace, symbol_name)
        == frame
    )


def test_cyclic_namespace_fails(empty_symbol_table: SymbolTable):
    """Test that adding a cyclic namespace raises a RuntimeError."""
    namespace_a = Identifier("namespace_a")
    namespace_b = Identifier("namespace_b")

    empty_symbol_table.add_namespace(namespace_a, namespace_b)
    empty_symbol_table.add_namespace(namespace_b, namespace_a)

    with pytest.raises(RuntimeError):
        empty_symbol_table.is_symbol_defined_in_namespace(
            namespace_a, Identifier("some_symbol")
        )


def test_update_namespaces():
    """Test that namespaces can be merged."""
    symbol_table_1 = SymbolTable()
    symbol_table_2 = SymbolTable()
    namespace = Identifier("shared_namespace")
    symbol_name = Identifier("shared_symbol")
    frame = ImportSymbolTableFrame(symbol_name)

    symbol_table_1.add_namespace(namespace)
    symbol_table_1.add_symbol(namespace, symbol_name, frame)

    symbol_table_2.update_namespaces(symbol_table_1)

    assert symbol_table_2.is_namespace_defined(namespace)
    assert symbol_table_2.is_symbol_defined_in_namespace(namespace, symbol_name)
    assert symbol_table_2.get_frame_from_namespace(namespace, symbol_name) == frame



"""Bound numerical parameters."""

__all__ = ["BoundIntParam", "BoundNatParam"]

from typing import Any, Iterable, Optional, Tuple

from fhy_core.constraint import Constraint, EquationConstraint
from fhy_core.expression import (
    BinaryExpression,
    BinaryOperation,
    Expression,
    IdentifierExpression,
    LiteralExpression,
)
from fhy_core.identifier import Identifier
from fhy_core.utils import Self

from .core import IntParam, Param
from .fundamental import NatParam


def _is_valid_bound_expression(expression: Expression) -> bool:
    if not isinstance(expression, BinaryExpression):
        return False
    if expression.operation not in (
        BinaryOperation.GREATER_EQUAL,
        BinaryOperation.GREATER,
        BinaryOperation.LESS_EQUAL,
        BinaryOperation.LESS,
    ):
        return False
    if not (
        (
            isinstance(expression.left, IdentifierExpression)
            or isinstance(expression.right, IdentifierExpression)
        )
        and (
            isinstance(expression.left, LiteralExpression)
            or isinstance(expression.right, LiteralExpression)
        )
    ):
        return False
    literal_expression: LiteralExpression
    if isinstance(expression.left, LiteralExpression):
        literal_expression = expression.left
    elif isinstance(expression.right, LiteralExpression):
        literal_expression = expression.right
    else:
        raise RuntimeError(
            "Somehow failed to find LiteralExpression in bound expression."
        )
    return isinstance(literal_expression.value, int)


def _invert_binary_comparison_operation(op: BinaryOperation) -> BinaryOperation:
    if op == BinaryOperation.GREATER:
        return BinaryOperation.LESS
    elif op == BinaryOperation.GREATER_EQUAL:
        return BinaryOperation.LESS_EQUAL
    elif op == BinaryOperation.LESS:
        return BinaryOperation.GREATER
    elif op == BinaryOperation.LESS_EQUAL:
        return BinaryOperation.GREATER_EQUAL
    else:
        raise ValueError(f"Cannot invert non-comparison binary operation: {op}")


def _get_bound_from_expression(
    literal_expression: LiteralExpression, op: BinaryOperation
) -> Tuple[bool, int, bool]:
    value = literal_expression.value
    if not isinstance(value, int):
        raise RuntimeError(
            "Somehow failed to find integer LiteralExpression in bound expression."
        )
    return (
        True
        if op in (BinaryOperation.GREATER, BinaryOperation.GREATER_EQUAL)
        else False,
        value,
        True
        if op in (BinaryOperation.GREATER_EQUAL, BinaryOperation.LESS_EQUAL)
        else False,
    )


class BoundIntParam(IntParam):
    """An interval integer parameter that supports basic arithmetic.

    Results can be represented as inclusive or exclusive via prefer_inclusive.
    """

    _prefer_inclusive: bool

    def __init__(
        self,
        name: Identifier | None = None,
        prefer_inclusive: bool = True,
        **kwargs: Any,
    ) -> None:
        super().__init__(name)
        self._prefer_inclusive = prefer_inclusive

    def add_constraint(self, constraint: Constraint) -> None:
        if not isinstance(constraint, EquationConstraint):
            raise TypeError(
                "BoundIntParam only supports EquationConstraint constraints."
            )
        if not _is_valid_bound_expression(constraint.convert_to_expression()):
            raise ValueError(
                "BoundIntParam only supports bound expressions of the form "
                '"x >= k", "x > k", "x <= k", or "x < k" where k is an integer.'
            )
        return super().add_constraint(constraint)

    def _iter_bounds(self) -> Iterable[Tuple[bool, int, bool]]:
        """Yield (is_lower, bound_value, inclusive) for constraints.

        Fail if any constraint is not a valid bound expression on the same
        variable.
        """
        for constraint in self._constraints:
            if not isinstance(constraint, EquationConstraint):
                raise RuntimeError(
                    "BoundIntParam somehow has non-EquationConstraint constraint: "
                    f"{type(constraint)}"
                )
            if not _is_valid_bound_expression(constraint.convert_to_expression()):
                raise RuntimeError(
                    "BoundIntParam somehow has non-bound expression constraint: "
                    f"{repr(constraint)}"
                )

            expression = constraint.convert_to_expression()
            identifier_expression: IdentifierExpression | None = None

            def set_or_check_identifier_expr(e: IdentifierExpression) -> None:
                nonlocal identifier_expression
                if identifier_expression is None:
                    identifier_expression = e
                elif identifier_expression != e:
                    raise RuntimeError(
                        "BoundIntParam somehow has bound constraints on "
                        "different variables."
                    )

            if not isinstance(expression, BinaryExpression):
                raise RuntimeError(
                    "Somehow bound constraint expression is not a BinaryExpression."
                )
            if isinstance(expression.left, IdentifierExpression):
                if not isinstance(expression.right, LiteralExpression):
                    raise RuntimeError(
                        "Somehow bound expression is not in the expected form."
                    )
                set_or_check_identifier_expr(expression.left)
                yield _get_bound_from_expression(expression.right, expression.operation)
            elif isinstance(expression.right, IdentifierExpression):
                if not isinstance(expression.left, LiteralExpression):
                    raise RuntimeError(
                        "Somehow bound expression is not in the expected form."
                    )
                set_or_check_identifier_expr(expression.right)
                yield _get_bound_from_expression(
                    expression.left,
                    _invert_binary_comparison_operation(expression.operation),
                )
            else:
                raise RuntimeError(
                    "Somehow bound expression is not in the expected form."
                )

    def _get_effective_min_max(self) -> Tuple[int | None, int | None]:
        """Return (min_int, max_int) represented by constraints.

        Semantics:
          x > k  => min_int >= k+1
          x >= k => min_int >= k
          x < k  => max_int <= k-1
          x <= k => max_int <= k
        None means unbounded.
        """
        min_int: int | None = None
        max_int: int | None = None

        for is_lower, bound, inclusive in self._iter_bounds():
            if is_lower:
                eff = bound if inclusive else bound + 1
                min_int = eff if min_int is None else max(min_int, eff)
            else:
                eff = bound if inclusive else bound - 1
                max_int = eff if max_int is None else min(max_int, eff)

        if min_int is not None and max_int is not None and min_int > max_int:
            raise ValueError(
                "Empty integer interval represented by constraints for "
                f"{self.variable}."
            )

        return min_int, max_int

    @staticmethod
    def _create_param_from_min_max(
        param: "BoundIntParam",
        min_int: Optional[int],
        max_int: Optional[int],
    ) -> "BoundIntParam":
        """Create a new BoundIntParam from min/max integers.

        Emits either inclusive or exclusive style depending on param.prefer_inclusive.
        """
        out = BoundIntParam(
            name=param.variable, prefer_inclusive=param._prefer_inclusive
        )
        if min_int is not None:
            if out._prefer_inclusive:
                out.add_lower_bound_constraint(min_int, True)
            else:
                out.add_lower_bound_constraint(min_int - 1, False)

        if max_int is not None:
            if out._prefer_inclusive:
                out.add_upper_bound_constraint(max_int, True)
            else:
                out.add_upper_bound_constraint(max_int + 1, False)
        return out

    @classmethod
    def between(
        cls: type[Self],
        lower_bound: int,
        upper_bound: int,
        name: Identifier | None = None,
        is_lower_inclusive: bool = True,
        is_upper_inclusive: bool = True,
        prefer_inclusive: bool = True,
    ) -> "BoundIntParam":
        p = cls(name, prefer_inclusive=prefer_inclusive)
        p.add_lower_bound_constraint(lower_bound, is_lower_inclusive)
        p.add_upper_bound_constraint(upper_bound, is_upper_inclusive)
        p._get_effective_min_max()
        return p

    @classmethod
    def with_lower_bound(
        cls: type[Self],
        lower_bound: int,
        name: Identifier | None = None,
        is_inclusive: bool = True,
        prefer_inclusive: bool = True,
    ) -> "BoundIntParam":
        p = cls(name, prefer_inclusive=prefer_inclusive)
        p.add_lower_bound_constraint(lower_bound, is_inclusive)
        p._get_effective_min_max()
        return p

    @classmethod
    def with_upper_bound(
        cls: type[Self],
        upper_bound: int,
        name: Identifier | None = None,
        is_inclusive: bool = True,
        prefer_inclusive: bool = True,
    ) -> "BoundIntParam":
        p = cls(name, prefer_inclusive=prefer_inclusive)
        p.add_upper_bound_constraint(upper_bound, is_inclusive)
        p._get_effective_min_max()
        return p

    @classmethod
    def exactly(
        cls: type[Self],
        value: int,
        name: Identifier | None = None,
        prefer_inclusive: bool = True,
    ) -> "BoundIntParam":
        p = cls(name, prefer_inclusive=prefer_inclusive)
        p.add_lower_bound_constraint(value, is_inclusive=True)
        p.add_upper_bound_constraint(value, is_inclusive=True)
        return p

    def _coerce_other(self, other: Any) -> "BoundIntParam":
        if isinstance(other, int):
            return BoundIntParam.exactly(other, prefer_inclusive=self._prefer_inclusive)
        elif isinstance(other, BoundIntParam):
            return other
        elif isinstance(other, IntParam):
            for other_constraint in other._constraints:
                if not _is_valid_bound_expression(
                    other_constraint.convert_to_expression()
                ):
                    raise TypeError(
                        "Cannot coerce IntParam with non-bound constraints "
                        "to BoundIntParam."
                    )
            wrapper_param = BoundIntParam(
                other.variable, prefer_inclusive=self._prefer_inclusive
            )
            wrapper_param._value = other._value
            Param.copy_constraints_to_new_param(other, wrapper_param)
            return wrapper_param
        raise TypeError(f"Unsupported operand type: {type(other)}")

    def __add__(self, other: Any) -> "BoundIntParam":
        coerced_other = self._coerce_other(other)
        if self.is_value_set() and coerced_other.is_value_set():
            return BoundIntParam.with_value(
                self.get_value() + coerced_other.get_value()
            )

        self_min, self_max = self._get_effective_min_max()
        other_min, other_max = coerced_other._get_effective_min_max()
        new_min = (
            None if (self_min is None or other_min is None) else self_min + other_min
        )
        new_max = (
            None if (self_max is None or other_max is None) else self_max + other_max
        )
        return self._create_param_from_min_max(self, new_min, new_max)

    def __radd__(self, other: Any) -> "BoundIntParam":
        return self.__add__(other)

    def __sub__(self, other: Any) -> "BoundIntParam":
        coerced_other = self._coerce_other(other)
        if self.is_value_set() and coerced_other.is_value_set():
            return BoundIntParam.with_value(
                self.get_value() - coerced_other.get_value()
            )

        self_min, self_max = self._get_effective_min_max()
        other_min, other_max = coerced_other._get_effective_min_max()
        new_min = (
            None if (self_min is None or other_max is None) else self_min - other_max
        )
        new_max = (
            None if (self_max is None or other_min is None) else self_max - other_min
        )
        return self._create_param_from_min_max(self, new_min, new_max)

    def __rsub__(self, other: Any) -> "BoundIntParam":
        return self._coerce_other(other).__sub__(self)

    def __neg__(self) -> "BoundIntParam":
        if self.is_value_set():
            return BoundIntParam.with_value(-self.get_value())

        self_min, self_max = self._get_effective_min_max()
        new_min = None if self_max is None else -self_max
        new_max = None if self_min is None else -self_min
        return self._create_param_from_min_max(self, new_min, new_max)


class BoundNatParam(BoundIntParam, NatParam):
    """A bounded natural number parameter (i.e., non-negative integers)."""

    def __init__(
        self,
        name: Identifier | None = None,
        is_zero_included: bool = True,
        prefer_inclusive: bool = True,
    ) -> None:
        super().__init__(
            name=name,
            is_zero_included=is_zero_included,
            prefer_inclusive=prefer_inclusive,
        )



"""Tests the identifier."""

from copy import copy
from unittest.mock import patch

from fhy_core.identifier import Identifier


def test_identifier_initialization():
    """Test that the identifier is initialized correctly."""
    identifier = Identifier("test_name")
    assert identifier.name_hint == "test_name"
    assert isinstance(identifier.id, int)


@patch.object(Identifier, "_next_id", 0)
def test_unique_id_generation():
    """Test that the identifier generates a unique ID."""
    id1 = Identifier("")
    id2 = Identifier("")
    id3 = Identifier("")

    assert id1.id == 0
    assert id2.id == 1
    assert id3.id == 2


def test_equality():
    """Test that the identifier equality is based on the ID."""
    id1 = Identifier("name")
    id2 = Identifier("name")
    id3 = Identifier("bar")

    assert id1 == id1  # noqa: PLR0124
    assert id1 != id2
    assert id2 != id3


def test_copy():
    """Test that the identifier can be copied."""
    identifier = Identifier("copy_test")
    copy_identifier = copy(identifier)

    assert copy_identifier.id == identifier.id
    assert copy_identifier.name_hint == identifier.name_hint
    assert copy_identifier == identifier


def test_string_representation():
    """Test that the identifier string representation is correct."""
    identifier = Identifier("test_repr")
    assert str(identifier) == "test_repr"
    assert repr(identifier) == f"test_repr::{identifier.id}"


def test_hash():
    """Test that the identifier hash is based on the ID."""
    identifier = Identifier("hash_test")
    assert hash(identifier) == hash(identifier.id)



"""Expression tree visitor and transformer."""

__all__ = [
    "ExpressionBasePass",
    "ExpressionVisitor",
    "ExpressionTransformer",
]

from abc import ABC
from typing import Any

from .core import (
    BinaryExpression,
    Expression,
    IdentifierExpression,
    LiteralExpression,
    UnaryExpression,
)


class ExpressionBasePass(ABC):
    """Base class for expression tree passes."""

    def __call__(self, expression: Expression) -> Any:
        return self.visit(expression)

    def visit(self, expression: Expression) -> Any:
        """Visit an expression.

        Args:
            expression: Expression to visit.

        Returns:
            Result of the visit.

        Raises:
            NotImplementedError: If the expression type is not supported.

        """
        match expression:
            case UnaryExpression():
                return self.visit_unary_expression(expression)
            case BinaryExpression():
                return self.visit_binary_expression(expression)
            case IdentifierExpression():
                return self.visit_identifier_expression(expression)
            case LiteralExpression():
                return self.visit_literal_expression(expression)
            case _:
                raise NotImplementedError(
                    f"Unsupported expression type: {type(expression)}"
                )

    def visit_unary_expression(self, unary_expression: UnaryExpression) -> Any:
        """Visit a unary expression.

        Args:
            unary_expression: Unary expression to visit.

        Returns:
            Result of the visit.

        """

    def visit_binary_expression(self, binary_expression: BinaryExpression) -> Any:
        """Visit a binary expression.

        Args:
            binary_expression: Binary expression to visit.

        Returns:
            Result of the visit.

        """

    def visit_identifier_expression(
        self, identifier_expression: IdentifierExpression
    ) -> Any:
        """Visit an identifier.

        Args:
            identifier_expression: Identifier expression to visit.

        Returns:
            Result of the visit.

        """

    def visit_literal_expression(self, literal_expression: LiteralExpression) -> Any:
        """Visit a literal.

        Args:
            literal_expression: Literal expression to visit.

        Returns:
            Result of the visit.

        """


class ExpressionVisitor(ExpressionBasePass, ABC):
    """Visitor for expression trees."""

    def __call__(self, expression: Expression) -> None:
        super().__call__(expression)

    def visit_unary_expression(self, unary_expression: UnaryExpression) -> None:
        self.visit(unary_expression.operand)

    def visit_binary_expression(self, binary_expression: BinaryExpression) -> None:
        self.visit(binary_expression.left)
        self.visit(binary_expression.right)

    def visit_identifier_expression(
        self, identifier_expression: IdentifierExpression
    ) -> None: ...

    def visit_literal_expression(
        self, literal_expression: LiteralExpression
    ) -> None: ...


class ExpressionTransformer(ExpressionBasePass, ABC):
    """Transformer for expression trees."""

    def __call__(self, expression: Expression) -> Expression:
        transformed_expression = super().__call__(expression)
        if not isinstance(transformed_expression, Expression):
            raise TypeError(
                f"Invalid transformed expression type: {type(transformed_expression)}"
            )
        return transformed_expression

    def visit_unary_expression(self, unary_expression: UnaryExpression) -> Expression:
        new_expression = self.visit(unary_expression.operand)
        return UnaryExpression(
            operation=unary_expression.operation, operand=new_expression
        )

    def visit_binary_expression(
        self, binary_expression: BinaryExpression
    ) -> Expression:
        new_left = self.visit(binary_expression.left)
        new_right = self.visit(binary_expression.right)
        return BinaryExpression(
            operation=binary_expression.operation, left=new_left, right=new_right
        )

    def visit_identifier_expression(
        self, identifier_expression: IdentifierExpression
    ) -> Expression:
        return IdentifierExpression(identifier=identifier_expression.identifier)

    def visit_literal_expression(
        self, literal_expression: LiteralExpression
    ) -> Expression:
        return LiteralExpression(value=literal_expression.value)



"""Fundamental parameter classes."""

__all__ = ["NatParam"]

from typing import Any

from fhy_core.constraint import EquationConstraint
from fhy_core.expression import (
    BinaryExpression,
    BinaryOperation,
    IdentifierExpression,
    LiteralExpression,
)
from fhy_core.identifier import Identifier

from .core import IntParam


class NatParam(IntParam):
    """Natural number parameter."""

    _is_zero_included: bool

    def __init__(
        self,
        name: Identifier | None = None,
        is_zero_included: bool = True,
        **kwargs: Any,
    ) -> None:
        super().__init__(name)
        self._is_zero_included = is_zero_included
        if self._is_zero_included:
            self.add_constraint(
                EquationConstraint(
                    self.variable,
                    BinaryExpression(
                        BinaryOperation.GREATER_EQUAL,
                        IdentifierExpression(self.variable),
                        LiteralExpression(0),
                    ),
                )
            )
        else:
            self.add_constraint(
                EquationConstraint(
                    self.variable,
                    BinaryExpression(
                        BinaryOperation.GREATER,
                        IdentifierExpression(self.variable),
                        LiteralExpression(0),
                    ),
                )
            )

    def add_lower_bound_constraint(
        self, lower_bound: int, is_inclusive: bool = True
    ) -> None:
        if self._is_zero_included:
            if lower_bound < 0:
                raise ValueError("Lower bound must be non-negative.")
            if not is_inclusive and lower_bound < 1:
                raise ValueError(
                    "Lower bound must be at least 1 if zero is included and "
                    "bound is exclusive."
                )
        elif is_inclusive:
            if lower_bound < 1:
                raise ValueError(
                    "Lower bound must be at least 1 when zero is not included."
                )
        elif lower_bound < 0:
            raise ValueError(
                "Lower bound must be non-negative when zero is not included "
                "and bound is exclusive."
            )

        return super().add_lower_bound_constraint(lower_bound, is_inclusive)

    def add_upper_bound_constraint(
        self, upper_bound: int, is_inclusive: bool = True
    ) -> None:
        if self._is_zero_included:
            if is_inclusive:
                if upper_bound < 0:
                    raise ValueError(
                        "Upper bound must be non-negative when zero is included."
                    )
            elif upper_bound < 1:
                raise ValueError(
                    "Upper bound must be at least 1 if zero is included and "
                    "bound is exclusive."
                )
        elif is_inclusive:
            if upper_bound < 1:
                raise ValueError(
                    "Upper bound must be at least 1 when zero is not included."
                )
        elif upper_bound < 2:  # noqa: PLR2004
            raise ValueError(
                "Upper bound must be at least 2 when zero is not included "
                "and bound is exclusive."
            )

        return super().add_upper_bound_constraint(upper_bound, is_inclusive)



"""Tests the core type system."""

import pytest
from fhy_core.types import (
    CoreDataType,
    TypeQualifier,
    get_core_data_type_bit_width,
    promote_core_data_types,
    promote_type_qualifiers,
)


@pytest.mark.parametrize(
    "core_data_type, expected_bit_width",
    [
        (CoreDataType.UINT8, 8),
        (CoreDataType.UINT16, 16),
        (CoreDataType.UINT32, 32),
        (CoreDataType.UINT64, 64),
        (CoreDataType.INT8, 8),
        (CoreDataType.INT16, 16),
        (CoreDataType.INT32, 32),
        (CoreDataType.INT64, 64),
        (CoreDataType.FLOAT16, 16),
        (CoreDataType.FLOAT32, 32),
        (CoreDataType.FLOAT64, 64),
        (CoreDataType.COMPLEX32, 32),
        (CoreDataType.COMPLEX64, 64),
        (CoreDataType.COMPLEX128, 128),
    ],
)
def test_get_core_data_type_bit_width(core_data_type, expected_bit_width):
    """Test get_core_data_type_bit_width function with various core data types."""
    assert get_core_data_type_bit_width(core_data_type) == expected_bit_width


@pytest.mark.parametrize(
    ("core_data_type1", "core_data_type2", "expected_core_data_type"),
    [
        (CoreDataType.UINT8, CoreDataType.UINT8, CoreDataType.UINT8),
        (CoreDataType.UINT8, CoreDataType.UINT16, CoreDataType.UINT16),
        (CoreDataType.UINT16, CoreDataType.UINT8, CoreDataType.UINT16),
        (CoreDataType.INT32, CoreDataType.INT64, CoreDataType.INT64),
        (
            CoreDataType.FLOAT16,
            CoreDataType.FLOAT32,
            CoreDataType.FLOAT32,
        ),
        (
            CoreDataType.FLOAT64,
            CoreDataType.FLOAT16,
            CoreDataType.FLOAT64,
        ),
        (
            CoreDataType.COMPLEX32,
            CoreDataType.COMPLEX64,
            CoreDataType.COMPLEX64,
        ),
        (
            CoreDataType.FLOAT32,
            CoreDataType.COMPLEX32,
            CoreDataType.COMPLEX64,
        ),
    ],
)
def test_promote_primitive_data_type(
    core_data_type1, core_data_type2, expected_core_data_type
):
    """Test primitive data types are correctly promoted."""
    assert (
        promote_core_data_types(core_data_type1, core_data_type2)
        == expected_core_data_type
    ), (
        f"Expected the promotion of {core_data_type1} and {core_data_type2} "
        f"to be {expected_core_data_type}."
    )


@pytest.mark.parametrize(
    ("type_qualifer1", "type_qualifer2", "expected_type_qualifer"),
    [
        (TypeQualifier.INPUT, TypeQualifier.INPUT, TypeQualifier.TEMP),
        (TypeQualifier.STATE, TypeQualifier.PARAM, TypeQualifier.TEMP),
        (TypeQualifier.PARAM, TypeQualifier.TEMP, TypeQualifier.TEMP),
        (TypeQualifier.PARAM, TypeQualifier.PARAM, TypeQualifier.PARAM),
    ],
)
def test_promote_type_qualifiers(
    type_qualifer1, type_qualifer2, expected_type_qualifer
):
    """Test type qualifiers are correctly promoted."""
    assert (
        promote_type_qualifiers(type_qualifer1, type_qualifer2)
        == expected_type_qualifer
    ), (
        f"Expected the promotion of {type_qualifer1} and {type_qualifer2} "
        f"to be {expected_type_qualifer}."
    )



"""Core symbol table."""

__all__ = [
    "FunctionSymbolTableFrame",
    "ImportSymbolTableFrame",
    "SymbolTable",
    "SymbolTableFrame",
    "VariableSymbolTableFrame",
]


from abc import ABC
from collections.abc import Callable
from dataclasses import dataclass, field
from typing import Generic, NoReturn, TypeVar

from fhy_core.identifier import Identifier

from .error import SymbolTableError
from .types import Type, TypeQualifier
from .utils import StrEnum


@dataclass(frozen=True)
class SymbolTableFrame(ABC):
    """Base symbol table frame."""

    name: Identifier


class ImportSymbolTableFrame(SymbolTableFrame):
    """Imported symbol frame."""


@dataclass(frozen=True)
class VariableSymbolTableFrame(SymbolTableFrame):
    """Variable symbol frame."""

    type: Type
    type_qualifier: TypeQualifier


class FunctionKeyword(StrEnum):
    """Function keyword."""

    PROCEDURE = "proc"
    OPERATION = "op"
    NATIVE = "native"


@dataclass(frozen=True)
class FunctionSymbolTableFrame(SymbolTableFrame):
    """Functions symbol frame."""

    keyword: FunctionKeyword
    signature: list[tuple[TypeQualifier, Type]] = field(default_factory=list)


_T = TypeVar("_T")
_SUCC_T = TypeVar("_SUCC_T")
_FAIL_T = TypeVar("_FAIL_T")


@dataclass(frozen=True)
class _SymbolTableSearchResult(Generic[_T]):
    value: _T


class SymbolTable:
    """Core nested symbol table comprised of various frames."""

    _table: dict[Identifier, dict[Identifier, SymbolTableFrame]]
    _parent_namespace: dict[Identifier, Identifier]

    def __init__(self) -> None:
        super().__init__()
        self._table = {}
        self._parent_namespace = {}

    def get_number_of_namespaces(self) -> int:
        """Return the number of namespaces in the symbol table."""
        return len(self._table)

    def add_namespace(
        self,
        namespace_name: Identifier,
        parent_namespace_name: Identifier | None = None,
    ) -> None:
        """Add a new namespace to the symbol table.

        Args:
            namespace_name: Name of the namespace to be added to the symbol table.
            parent_namespace_name: Name of the parent namespace.

        Raises:
            SymbolTableError: If the namespace is already defined in the symbol table.

        """
        if self.is_namespace_defined(namespace_name):
            raise SymbolTableError(
                f"Namespace {namespace_name} already defined in the symbol table."
            )
        self._table[namespace_name] = {}
        if parent_namespace_name:
            self._parent_namespace[namespace_name] = parent_namespace_name

    def is_namespace_defined(self, namespace_name: Identifier) -> bool:
        """Check if a namespace exists in the symbol table.

        Args:
            namespace_name: Name of the namespace to be checked.

        Returns:
            True if the namespace exists in the symbol table, False otherwise.

        """
        return namespace_name in self._table

    def get_namespace(
        self, namespace_name: Identifier
    ) -> dict[Identifier, SymbolTableFrame]:
        """Retrieve a namespace from the symbol table.

        Args:
            namespace_name: Name of the namespace to be retrieved from
                the symbol table.

        Returns:
            Namespace table.

        Raises:
            SymbolTableError: If the namespace does not exist in the symbol table.

        """
        if namespace_name not in self._table:
            raise SymbolTableError(
                f"Namespace {namespace_name} not found in the symbol table."
            )

        return self._table[namespace_name]

    def update_namespaces(self, other_symbol_table: "SymbolTable") -> None:
        """Update the symbol table with new namespaces from another symbol table.

        Args:
            other_symbol_table: Symbol table to update with.

        """
        self._table.update(other_symbol_table._table)
        self._parent_namespace.update(other_symbol_table._parent_namespace)

    def add_symbol(
        self,
        namespace_name: Identifier,
        symbol_name: Identifier,
        frame: SymbolTableFrame,
    ) -> None:
        """Add a symbol to the symbol table.

        Args:
            namespace_name: Name of the namespace to add the symbol to.
            symbol_name: Name of the symbol to be added.
            frame: Frame to be added to the symbol table.

        Raises:
            SymbolTableError: If the symbol is already defined in the namespace.

        """
        if self.is_symbol_defined_in_namespace(namespace_name, symbol_name):
            raise SymbolTableError(
                f"Symbol {symbol_name} already defined in namespace {namespace_name}."
            )
        self._table[namespace_name][symbol_name] = frame

    def is_symbol_defined(self, symbol_name: Identifier) -> bool:
        """Check if a symbol exists in the symbol table.

        Args:
            symbol_name: Name of the symbol to check.

        Returns:
            True if the symbol exists in the symbol table, False otherwise.

        """
        return self._search_table_with_action(
            lambda _, candidate_symbol_name: _SymbolTableSearchResult(True)
            if symbol_name == candidate_symbol_name
            else None,
            lambda: False,
        )

    def is_symbol_defined_in_namespace(
        self, namespace_name: Identifier, symbol_name: Identifier
    ) -> bool:
        """Check if a symbol exists in a specific namespace.

        Args:
            namespace_name: Name of the namespace to check.
            symbol_name: Name of the symbol to check.

        Returns:
            True if the symbol exists in the namespace, False otherwise.

        """

        def is_symbol_defined_in_namespace(
            namespace_name: Identifier,
        ) -> _SymbolTableSearchResult[bool] | None:
            if symbol_name in self._table[namespace_name]:
                return _SymbolTableSearchResult(True)
            else:
                return None

        return self._search_namespace_with_action(
            namespace_name, is_symbol_defined_in_namespace, lambda: False
        )

    def get_frame(self, symbol_name: Identifier) -> SymbolTableFrame:
        """Retrieve a frame from the symbol table.

        Args:
            symbol_name: Name of the symbol to retrieve the frame for.

        Returns:
            The frame for the given symbol.

        Raises:
            SymbolTableError: If the symbol is not found in the symbol table or
                there is a cyclic namespace dependency.

        """

        def get_frame(
            namespace_name: Identifier, candidate_symbol_name: Identifier
        ) -> _SymbolTableSearchResult[SymbolTableFrame] | None:
            if symbol_name == candidate_symbol_name:
                return _SymbolTableSearchResult(
                    self._table[namespace_name][symbol_name]
                )
            else:
                return None

        def raise_symbol_not_found() -> NoReturn:
            raise SymbolTableError(
                f"Symbol {symbol_name} not found in the symbol table."
            )

        return self._search_table_with_action(get_frame, raise_symbol_not_found)

    def get_frame_from_namespace(
        self, namespace_name: Identifier, symbol_name: Identifier
    ) -> SymbolTableFrame:
        """Retrieve a frame from a specific namespace.

        Args:
            namespace_name: Name of the namespace.
            symbol_name: Name of the symbol to retrieve the frame for.

        Returns:
            The frame for the given symbol in the given namespace.

        Raises:
            SymbolTableError: If the symbol is not found in the namespace or
                there is a cyclic namespace dependency.

        """

        def get_frame_in_namespace(
            namespace_name: Identifier,
        ) -> _SymbolTableSearchResult[SymbolTableFrame] | None:
            if symbol_name in self._table[namespace_name]:
                return _SymbolTableSearchResult(
                    self._table[namespace_name][symbol_name]
                )
            else:
                return None

        def raise_symbol_not_found() -> NoReturn:
            raise SymbolTableError(
                f"Symbol {symbol_name} not found in namespace {namespace_name}."
            )

        return self._search_namespace_with_action(
            namespace_name, get_frame_in_namespace, raise_symbol_not_found
        )

    def _search_table_with_action(
        self,
        action: Callable[
            [Identifier, Identifier], _SymbolTableSearchResult[_SUCC_T] | None
        ],
        action_fail_func: Callable[[], _FAIL_T],
    ) -> _SUCC_T | _FAIL_T:
        for namespace_name, namespace_table in self._table.items():
            for symbol_name in namespace_table:
                result = action(namespace_name, symbol_name)
                if result is not None:
                    return result.value

        return action_fail_func()

    def _search_namespace_with_action(
        self,
        namespace_name: Identifier,
        action: Callable[[Identifier], _SymbolTableSearchResult[_SUCC_T] | None],
        action_fail_func: Callable[[], _FAIL_T],
    ) -> _SUCC_T | _FAIL_T:
        if not self.is_namespace_defined(namespace_name):
            raise SymbolTableError(
                f"Namespace {namespace_name} not found in the symbol table."
            )

        current_namespace_name: Identifier | None = namespace_name
        seen_namespace_names = set()
        while current_namespace_name is not None:
            if current_namespace_name in seen_namespace_names:
                raise RuntimeError(f"Namespace {current_namespace_name} is cyclic.")
            seen_namespace_names.add(current_namespace_name)

            result = action(current_namespace_name)
            if result is not None:
                return result.value

            current_namespace_name = self._parent_namespace.get(
                current_namespace_name, None
            )

        return action_fail_func()



"""Unique identifier for named compiler objects."""

__all__ = ["Identifier"]

from typing import Any


class Identifier:
    """Unique name."""

    _next_id: int = 0
    _id: int
    _name_hint: str

    def __init__(self, name_hint: str) -> None:
        # TODO: Add a lock to implement RMW atomicity for _next_id.
        self._id = Identifier._next_id
        Identifier._next_id += 1
        self._name_hint = name_hint

    @property
    def name_hint(self) -> str:
        return self._name_hint

    @property
    def id(self) -> int:
        return self._id

    def __copy__(self) -> "Identifier":
        identifier = Identifier.__new__(Identifier)
        identifier._id = self._id
        identifier._name_hint = self._name_hint

        return identifier

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, Identifier) and self._id == other._id

    def __hash__(self) -> int:
        return hash(self._id)

    def __str__(self) -> str:
        return self._name_hint

    def __repr__(self) -> str:
        return f"{self._name_hint}::{self._id}"



"""FhY compiler core utilities."""

__version__ = "0.0.3"

__all__ = [
    "Constraint",
    "EquationConstraint",
    "InSetConstraint",
    "NotInSetConstraint",
    "FhYCoreTypeError",
    "SymbolTableError",
    "register_error",
    "BinaryExpression",
    "BinaryOperation",
    "Expression",
    "ExpressionBasePass",
    "ExpressionVisitor",
    "IdentifierExpression",
    "LiteralExpression",
    "SymbolType",
    "UnaryExpression",
    "UnaryOperation",
    "collect_identifiers",
    "copy_expression",
    "is_satisfiable",
    "parse_expression",
    "pformat_expression",
    "replace_identifiers",
    "simplify_expression",
    "substitute_identifiers",
    "Identifier",
    "CategoricalParam",
    "IntParam",
    "BoundIntParam",
    "NatParam",
    "BoundNatParam",
    "OrdinalParam",
    "Param",
    "PermParam",
    "RealParam",
    "FunctionKeyword",
    "FunctionSymbolTableFrame",
    "ImportSymbolTableFrame",
    "SymbolTable",
    "SymbolTableFrame",
    "VariableSymbolTableFrame",
    "CoreDataType",
    "DataType",
    "IndexType",
    "NumericalType",
    "PrimitiveDataType",
    "TemplateDataType",
    "TupleType",
    "Type",
    "TypeQualifier",
    "get_core_data_type_bit_width",
    "promote_core_data_types",
    "promote_primitive_data_types",
    "promote_type_qualifiers",
    "IntEnum",
    "Lattice",
    "PartiallyOrderedSet",
    "Stack",
    "StrEnum",
    "format_comma_separated_list",
    "get_array_size_in_bits",
    "invert_dict",
    "invert_frozen_dict",
]


from .constraint import (
    Constraint,
    EquationConstraint,
    InSetConstraint,
    NotInSetConstraint,
)
from .error import FhYCoreTypeError, SymbolTableError, register_error
from .expression import (
    BinaryExpression,
    BinaryOperation,
    Expression,
    ExpressionBasePass,
    ExpressionVisitor,
    IdentifierExpression,
    LiteralExpression,
    SymbolType,
    UnaryExpression,
    UnaryOperation,
    collect_identifiers,
    copy_expression,
    is_satisfiable,
    parse_expression,
    pformat_expression,
    replace_identifiers,
    simplify_expression,
    substitute_identifiers,
)
from .identifier import Identifier
from .param import (
    BoundIntParam,
    BoundNatParam,
    CategoricalParam,
    IntParam,
    NatParam,
    OrdinalParam,
    Param,
    PermParam,
    RealParam,
)
from .symbol_table import (
    FunctionKeyword,
    FunctionSymbolTableFrame,
    ImportSymbolTableFrame,
    SymbolTable,
    SymbolTableFrame,
    VariableSymbolTableFrame,
)
from .types import (
    CoreDataType,
    DataType,
    IndexType,
    NumericalType,
    PrimitiveDataType,
    TemplateDataType,
    TupleType,
    Type,
    TypeQualifier,
    get_core_data_type_bit_width,
    promote_core_data_types,
    promote_primitive_data_types,
    promote_type_qualifiers,
)
from .utils import (
    IntEnum,
    Lattice,
    PartiallyOrderedSet,
    Stack,
    StrEnum,
    format_comma_separated_list,
    get_array_size_in_bits,
    invert_dict,
    invert_frozen_dict,
)



"""Tests the constraint utility."""

import pytest
from fhy_core.constraint import EquationConstraint, InSetConstraint, NotInSetConstraint
from fhy_core.expression import (
    BinaryExpression,
    BinaryOperation,
    IdentifierExpression,
    LiteralExpression,
    LiteralType,
    UnaryExpression,
    UnaryOperation,
)
from fhy_core.identifier import Identifier

from .utils import assert_exact_expression_equality, mock_identifier


@pytest.mark.parametrize(
    "constraint, value, expected_outcome",
    [
        (
            EquationConstraint(mock_identifier("x", 0), LiteralExpression(True)),
            LiteralExpression(0),
            True,
        ),
        (
            EquationConstraint(mock_identifier("x", 0), LiteralExpression(False)),
            LiteralExpression(0),
            False,
        ),
        (
            EquationConstraint(
                mock_identifier("x", 0), IdentifierExpression(mock_identifier("x", 0))
            ),
            LiteralExpression(True),
            True,
        ),
        (
            EquationConstraint(
                mock_identifier("x", 0), IdentifierExpression(mock_identifier("x", 0))
            ),
            LiteralExpression(False),
            False,
        ),
        (
            EquationConstraint(
                mock_identifier("x", 0),
                UnaryExpression(UnaryOperation.LOGICAL_NOT, LiteralExpression(True)),
            ),
            LiteralExpression(True),
            False,
        ),
        (
            EquationConstraint(
                mock_identifier("x", 0),
                UnaryExpression(UnaryOperation.LOGICAL_NOT, LiteralExpression(False)),
            ),
            LiteralExpression(True),
            True,
        ),
        (
            EquationConstraint(
                mock_identifier("x", 0),
                BinaryExpression(
                    BinaryOperation.LOGICAL_AND,
                    LiteralExpression(True),
                    LiteralExpression(True),
                ),
            ),
            LiteralExpression(True),
            True,
        ),
        (
            EquationConstraint(
                mock_identifier("x", 0),
                BinaryExpression(
                    BinaryOperation.LOGICAL_AND,
                    LiteralExpression(True),
                    LiteralExpression(False),
                ),
            ),
            LiteralExpression(True),
            False,
        ),
        (
            EquationConstraint(
                mock_identifier("x", 0),
                BinaryExpression(
                    BinaryOperation.LOGICAL_OR,
                    LiteralExpression(True),
                    LiteralExpression(False),
                ),
            ),
            LiteralExpression(0),
            True,
        ),
        (
            EquationConstraint(
                mock_identifier("x", 0),
                BinaryExpression(
                    BinaryOperation.LOGICAL_OR,
                    LiteralExpression(False),
                    LiteralExpression(False),
                ),
            ),
            LiteralExpression(0),
            False,
        ),
        (
            EquationConstraint(
                mock_identifier("x", 0),
                BinaryExpression(
                    BinaryOperation.EQUAL,
                    LiteralExpression(True),
                    LiteralExpression(True),
                ),
            ),
            LiteralExpression(0),
            True,
        ),
        (
            EquationConstraint(
                mock_identifier("x", 0),
                BinaryExpression(
                    BinaryOperation.EQUAL,
                    LiteralExpression(True),
                    LiteralExpression(False),
                ),
            ),
            LiteralExpression(0),
            False,
        ),
        (
            EquationConstraint(
                mock_identifier("x", 0),
                BinaryExpression(
                    BinaryOperation.NOT_EQUAL,
                    LiteralExpression(True),
                    LiteralExpression(True),
                ),
            ),
            LiteralExpression(0),
            False,
        ),
        (
            EquationConstraint(
                mock_identifier("x", 0),
                BinaryExpression(
                    BinaryOperation.NOT_EQUAL,
                    LiteralExpression(True),
                    LiteralExpression(False),
                ),
            ),
            LiteralExpression(0),
            True,
        ),
        (
            EquationConstraint(
                mock_identifier("x", 0),
                BinaryExpression(
                    BinaryOperation.LESS, LiteralExpression(5), LiteralExpression(10)
                ),
            ),
            LiteralExpression(0),
            True,
        ),
        (
            EquationConstraint(
                mock_identifier("x", 0),
                BinaryExpression(
                    BinaryOperation.LESS, LiteralExpression(10), LiteralExpression(5)
                ),
            ),
            LiteralExpression(0),
            False,
        ),
        (
            EquationConstraint(
                mock_identifier("x", 0),
                BinaryExpression(
                    BinaryOperation.LESS_EQUAL,
                    LiteralExpression(10),
                    LiteralExpression(10),
                ),
            ),
            LiteralExpression(0),
            True,
        ),
        (
            EquationConstraint(
                mock_identifier("x", 0),
                BinaryExpression(
                    BinaryOperation.LESS_EQUAL,
                    LiteralExpression(10),
                    LiteralExpression(5),
                ),
            ),
            LiteralExpression(0),
            False,
        ),
        (
            EquationConstraint(
                mock_identifier("x", 0),
                BinaryExpression(
                    BinaryOperation.GREATER, LiteralExpression(10), LiteralExpression(5)
                ),
            ),
            LiteralExpression(0),
            True,
        ),
        (
            EquationConstraint(
                mock_identifier("x", 0),
                BinaryExpression(
                    BinaryOperation.GREATER, LiteralExpression(5), LiteralExpression(10)
                ),
            ),
            LiteralExpression(0),
            False,
        ),
        (
            EquationConstraint(
                mock_identifier("x", 0),
                BinaryExpression(
                    BinaryOperation.GREATER_EQUAL,
                    LiteralExpression(10),
                    LiteralExpression(10),
                ),
            ),
            LiteralExpression(0),
            True,
        ),
        (
            EquationConstraint(
                mock_identifier("x", 0),
                BinaryExpression(
                    BinaryOperation.GREATER_EQUAL,
                    LiteralExpression(5),
                    LiteralExpression(10),
                ),
            ),
            LiteralExpression(0),
            False,
        ),
    ],
)
def test_equation_constraint_checks_correctly(
    constraint: EquationConstraint,
    value: LiteralExpression,
    expected_outcome: bool,
):
    """Test the equation constraint evaluates correctly when checked."""
    assert constraint.is_satisfied(value) == expected_outcome


@pytest.mark.parametrize(
    "constraint, value, expected_outcome",
    [
        (
            InSetConstraint(mock_identifier("x", 0), {1, 2, 3}),
            1,
            True,
        ),
        (
            InSetConstraint(mock_identifier("x", 0), {1, 2, 3}),
            4,
            False,
        ),
        (
            InSetConstraint(mock_identifier("x", 0), {"a", "b", "c"}),
            "a",
            True,
        ),
        (
            InSetConstraint(mock_identifier("y", 1), {"a", "b", "c"}),
            "d",
            False,
        ),
    ],
)
def test_in_set_constraint_checks_correctly(
    constraint: InSetConstraint,
    value: LiteralExpression,
    expected_outcome: bool,
):
    """Test the in-set constraint evaluates correctly when checked."""
    assert constraint.is_satisfied(value) == expected_outcome


@pytest.mark.parametrize(
    "constraint, values, expected_outcome",
    [
        (
            NotInSetConstraint(mock_identifier("x", 0), {1, 2, 3}),
            1,
            False,
        ),
        (
            NotInSetConstraint(mock_identifier("x", 0), {1, 2, 3}),
            4,
            True,
        ),
        (
            NotInSetConstraint(mock_identifier("x", 0), {"a", "b", "c"}),
            "a",
            False,
        ),
        (
            NotInSetConstraint(mock_identifier("y", 1), {"a", "b", "c"}),
            "d",
            True,
        ),
    ],
)
def test_not_in_set_constraint_checks_correctly(
    constraint: NotInSetConstraint,
    values: dict[Identifier, LiteralType],
    expected_outcome: bool,
):
    """Test the not-in-set constraint evaluates correctly when checked."""
    assert constraint.is_satisfied(values) == expected_outcome


# TODO: If there is ever an equality for constraints, use this instead of
#       accessing private attributes for the three following tests.
def test_copy_equation_constraint():
    """Test the equation constraint is copied correctly."""
    constraint = EquationConstraint(mock_identifier("x", 0), LiteralExpression(True))
    copy = constraint.copy()
    assert constraint.variable == copy.variable
    assert_exact_expression_equality(constraint._expression, copy._expression)
    assert constraint is not copy


def test_copy_in_set_constraint():
    """Test the in-set constraint is copied correctly."""
    constraint = InSetConstraint(mock_identifier("x", 0), {1, 2, 3})
    copy = constraint.copy()
    assert constraint.variable == copy.variable
    assert constraint._valid_values == copy._valid_values
    assert constraint is not copy


def test_copy_not_in_set_constraint():
    """Test the not-in-set constraint is copied correctly."""
    constraint = NotInSetConstraint(mock_identifier("x", 0), {1, 2, 3})
    copy = constraint.copy()
    assert constraint.variable == copy.variable
    assert constraint._invalid_values == copy._invalid_values
    assert constraint is not copy


def test_convert_equation_constraint_to_expression():
    """Test the equation constraint is converted to an expression correctly."""
    constraint_expression = BinaryExpression(
        BinaryOperation.EQUAL,
        IdentifierExpression(mock_identifier("x", 0)),
        LiteralExpression(True),
    )
    constraint = EquationConstraint(mock_identifier("x", 0), constraint_expression)
    expression = constraint.convert_to_expression()
    assert_exact_expression_equality(constraint_expression, expression)


def test_convert_in_set_constraint_to_expression():
    """Test the in-set constraint is converted to an expression correctly."""
    constraint = InSetConstraint(mock_identifier("x", 0), {1, 2})
    expression = constraint.convert_to_expression()
    expected_expression = BinaryExpression(
        BinaryOperation.LOGICAL_OR,
        BinaryExpression(
            BinaryOperation.EQUAL,
            IdentifierExpression(mock_identifier("x", 0)),
            LiteralExpression(1),
        ),
        BinaryExpression(
            BinaryOperation.EQUAL,
            IdentifierExpression(mock_identifier("x", 0)),
            LiteralExpression(2),
        ),
    )
    assert_exact_expression_equality(expected_expression, expression)


def test_convert_not_in_set_constraint_to_expression():
    """Test the not-in-set constraint is converted to an expression correctly."""
    constraint = NotInSetConstraint(mock_identifier("x", 0), {1, 2})
    expression = constraint.convert_to_expression()
    expected_expression = BinaryExpression(
        BinaryOperation.LOGICAL_AND,
        BinaryExpression(
            BinaryOperation.NOT_EQUAL,
            IdentifierExpression(mock_identifier("x", 0)),
            LiteralExpression(1),
        ),
        BinaryExpression(
            BinaryOperation.NOT_EQUAL,
            IdentifierExpression(mock_identifier("x", 0)),
            LiteralExpression(2),
        ),
    )
    assert_exact_expression_equality(expected_expression, expression)



"""Basic expression passes."""

__all__ = [
    "collect_identifiers",
    "copy_expression",
    "substitute_identifiers",
]

from fhy_core.expression.core import (
    Expression,
    IdentifierExpression,
)
from fhy_core.expression.visitor import ExpressionTransformer, ExpressionVisitor
from fhy_core.identifier import Identifier


class IdentifierCollector(ExpressionVisitor):
    """Collect all identifiers in an expression tree."""

    _identifiers: set[Identifier]

    def __init__(self) -> None:
        self._identifiers = set()

    @property
    def identifiers(self) -> set[Identifier]:
        return self._identifiers

    def visit_identifier_expression(
        self, identifier_expression: IdentifierExpression
    ) -> None:
        self._identifiers.add(identifier_expression.identifier)


def collect_identifiers(expression: Expression) -> set[Identifier]:
    """Collect all identifiers in an expression tree.

    Args:
        expression: Expression to collect identifiers from.

    Returns:
        Set of identifiers in the expression.

    """
    collector = IdentifierCollector()
    collector(expression)
    return collector.identifiers


class ExpressionCopier(ExpressionTransformer):
    """Shallow copier for an expression tree."""


def copy_expression(expression: Expression) -> Expression:
    """Shallow-copy an expression.

    Args:
        expression: Expression to copy.

    Returns:
        Copied expression.

    """
    return ExpressionCopier()(expression)


class IdentifierSubstituter(ExpressionTransformer):
    """Substitute identifiers in an expression tree."""

    _substitutions: dict[Identifier, Expression]

    def __init__(self, substitutions: dict[Identifier, Expression]) -> None:
        self._substitutions = substitutions

    def visit_identifier_expression(
        self, identifier_expression: IdentifierExpression
    ) -> Expression:
        identifier = identifier_expression.identifier
        if identifier in self._substitutions:
            return self._substitutions[identifier]
        return identifier_expression


def substitute_identifiers(
    expression: Expression, substitutions: dict[Identifier, Expression]
) -> Expression:
    """Substitute identifiers in an expression tree.

    Args:
        expression: Expression to substitute identifiers in.
        substitutions: Substitutions to make.

    Returns:
        Expression with identifiers substituted.

    """
    return IdentifierSubstituter(substitutions)(expression)


def replace_identifiers(
    expression: Expression, replacements: dict[Identifier, Identifier]
) -> Expression:
    """Replace identifiers in an expression tree.

    Args:
        expression: Expression to replace identifiers in.
        replacements: Replacements to make.

    Returns:
        Expression with identifiers replaced.

    """
    substitutions: dict[Identifier, Expression] = {
        old_identifier: IdentifierExpression(new_identifier)
        for old_identifier, new_identifier in replacements.items()
    }
    return substitute_identifiers(expression, substitutions)
